"""
ReBNF: Regexes for Extended Backus-Naur Form (EBNF)

Copyright (C) 2023 opsocket <opsocket@pm.me>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses>.
"""
from . import lexers
from . import parsers

assert all((lexers, parsers))

__author__ = "opsocket <opsocket@pm.me>"
__description__ = "Lexer and parser for the ReBNF metasyntax language"
__version__ = "0.7"
__ascii__ = """
ooooooooo.             oooooooooo.  ooooo      ooo oooooooooooo 
`888   `Y88.           `888'   `Y8b `888b.     `8' `888'     `8 
 888   .d88'  .ooooo.   888     888  8 `88b.    8   888         
 888ooo88P'  d88' `88b  888oooo888'  8   `88b.  8   888oooo8    
 888`88b.    888ooo888  888    `88b  8     `88b.8   888    "    
 888  `88b.  888    .o  888    .88P  8       `888   888         
o888o  o888o `Y8bod8P' o888bood8P'  o8o        `8  o888o       """

__doc__ = Rf"""

{__ascii__} v{__version__}

{__description__}

"""


def tokenize(code) -> list[str]:
    """
    Tokenizes the provided rebnf code using the REBNFLexer.
    
    :param      code:  The rebnf code to tokenize.
    :type       code:  str
    
    :returns:   A list of tokens representing the code.
    :rtype:     list[str]
    """
    lexer = lexers.REBNFLexer()
    return lexer.tokenize(code)


def parse(code) -> dict:
    """
    Parses the given code using the REBNFParser.
    
    :param      code:  The code to parse.
    :type       code:  str
    
    :returns:   The parsed dictionary
    :rtype:     dict
    """
    parser = parsers.REBNFParser()
    return parser.parse(code)



"""
ReBNF: Regexes for Extended Backus-Naur Form (EBNF)

Copyright (C) 2023 opsocket <opsocket@pm.me>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses>.
"""
from ply import lex
import unicodedata
import re


def get_unicode_name(character):
    # replace non-alphanumeric characters with underscores
    sc = re.sub(r"[^a-zA-Z0-9]", "_", unicodedata.name(character))
    # convert to uppercase
    sc = sc.upper()
    # remove leading and trailing underscores
    sc = sc.strip("_")
    # replace consecutive underscores with a single underscore
    sc = re.sub(r"_+", "_", sc)
    # return upper snake cased character name
    return sc


# --


class Token(lex.LexToken):
    """
    This class describes a simple token wrapper.
    @see REBNFLexer.tokenize
    """

    def __init__(self, token: lex.LexToken):
        assert isinstance(token, lex.LexToken)
        super().__init__()
        self.__dict__.update(token.__dict__)

    @property
    def lcpos(self):
        return (
            self.lineno,
            self.column,
        )

    def __str__(self):
        return f"{self.__class__.__name__}({self.type}, {repr(self.value)}, {self.lcpos}, {self.lexpos})"


# --

# def group(*choices): return '(' + '|'.join(choices) + ')'
# def maybe(*choices): return group(*choices) + '?'

# --

literal_type_map = {
    '>': 'GREATER',
    '{': 'LBRACE',
    '<': 'LESS',
    '(': 'LPAR',
    '[': 'LSQB',
    '-': 'MINUS',
    '+': 'PLUS',
    '}': 'RBRACE',
    ')': 'RPAR',
    ']': 'RSQB',
    ';': 'SEMICOLON',
    '*': 'ASTERISK',
    '|': 'VBAR',
    '?': 'OPTIONAL',
    '.': 'DOT',
    # "...": "ELLIPSIS",
}


class REBNFLexer(object):
    """
    This class describes a rebnf lexer.
    """

    literals = literal_type_map.keys()

    tokens = (
        "ASSIGN",
        "COMMENT",
        "NAME",
        "NEWLINE",
        "REGEX",
        "STRING",
        *literal_type_map.values(),
    )

    t_ASSIGN = r"::=|:|="
    t_ignore = " \t\r\f\b"

    def t_NEWLINE(self, t):
        r"\n"
        t.lexer.lineno += t.value.count("\n")
        return t

    def t_STRING(self, t):
        r"r?(\'[^\n\'\\]*(\\.[^\n\'\\]*)*\'|\"[^\n\"\\]*(\\.[^\n\"\\]*)*\")"
        if t.value.startswith("r"):
            t.type = "REGEX"
        return t

    def t_COMMENT(self, t):
        r"(\#.*\n?|\(\*(?:.|\n)*?\*\))"
        t.lexer.lineno += t.value.count("\n")
        return t

    @lex.TOKEN("|".join(re.escape(k) for k in literal_type_map.keys()))
    def t_LITERAL(self, t):
        t.type = literal_type_map.get(t.value)
        return t

    # @lex.TOKEN('|'.join([
    #     r"import\ +(\w+(?:\ +as\ +\w+)?)",
    #     r"from\ +(\.*\w+(?:\.\w+)*)\ +import\ +(\w+(?:\ +as\ +\w+)?(?:,\ *\w+(?:\ +as\ +\w+)?)*)\ *",
    #     r"from\ +(\.*\w+(?:\.\w+)*)\ +import\ +\(\s+(\w+(?:\ +as\ +\w+)?(?:,\s+\w+(?:\ +as\ +\w+)?)*)\ *",
    # ]))
    # def t_IMPORT(self, t):
    #     return t

    def t_NAME(self, t):
        r"<[a-zA-Z][a-zA-Z0-9_]*>|[a-zA-Z][a-zA-Z0-9_]*"
        # support for optional identifier enclosures
        if t.value.startswith("<") and t.value.endswith(">"):
            t.value = t.value[1:-1]
        return t

    # def t_SYMBOL(self, t):
    #     r"<|>|-|\n|\t|\r|\f|\b"
    #     return t

    def find_column(self, data, token):
        line_start = data.rfind("\n", 0, token.lexpos) + 1
        return (token.lexpos - line_start) + 1

    def t_error(self, t):
        print(
            f"\033[91mlexer error: at line {t.lineno}, column {self.find_column(t.lexer.lexdata, t)}, unexpected token {repr(t.value[0])}\033[0m"
        )
        exit(1)

    def __init__(self):
        self.lexer = lex.lex(module=self)

    def token(self):
        return self.lexer.token()

    def tokenize(self, code):
        self.lexer.input(code)
        output = list()
        while tok := self.lexer.token():
            tok.column = self.find_column(self.lexer.lexdata, tok)
            output.append(Token(tok))
        return output



"""
ReBNF: Regexes for Extended Backus-Naur Form (EBNF)

Copyright (C) 2023 opsocket <opsocket@pm.me>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses>.
"""
from . import lexers
from ply import yacc
import itertools


class REBNFParser:
    """
    This class describes a rebnf parser.
    """

    tokens = lexers.REBNFLexer.tokens

    def __init__(self):
        self.lexer = lexers.REBNFLexer()
        self.parser = yacc.yacc(module=self)
        # self.parser = yacc.yacc(module=self, debuglog=yacc.NullLogger())
        self.ast = []

    def p_start(self, p):
        """
        start : grammar start
              | grammar
        """
        p[0] = self.ast

    def p_grammar(self, p):
        '''
        grammar : comment
                | import
                | rule
                | newline
        '''
        if p[1]:
            self.ast += p[1],

    def p_newline(self, p):
        '''
        newline : NEWLINE newline
                | NEWLINE
        '''
        pass

    def p_comment(self, p):
        '''
        comment : COMMENT
        '''
        p[0] = ('COMMENT', p[1])

    # -- IMPORTS --

    def p_import(self, p):
        '''
        import : import_statement newline
               | import_statement
               | from_statement newline
               | from_statement
        '''
        p[0] = ('IMPORT', p[1])

    def p_import_statement(self, p):
        '''
        import_statement : NAME import_items
        '''
        if p[1] == 'import':
            p[0] = tuple(p[1:])

    def p_from_statement(self, p):
        '''
        from_statement : NAME dotted_name NAME ASTERISK
                       | NAME dotted_name NAME import_items
        '''
        if p[1] == "from" and p[3] == "import":
            p[0] = tuple(p[1:])

    def p_dotted_name(self, p):
        """
        dotted_name : dots name_parts
                    | name_parts
        """
        p[0] = "".join(p[1:])

    def p_dotted_name_dots(self, p):
        """
        dots : DOT dots
             | DOT
        """
        p[0] = "".join(p[1:])

    def p_dotted_name_name_parts(self, p):
        """
        name_parts : NAME DOT name_parts
                   | NAME
        """
        p[0] = "".join(p[1:])

    def p_import_items(self, p):
        '''
        import_items : import_name "," import_items
                     | import_name
        '''
        if len(p) == 3:
            p[0] = itertools.chain.from_iterable([p[1], p[2]])
        if len(p) == 2:
            p[0] = p[1]

    def p_import_name(self, p):
        '''
        import_name : NAME NAME NAME
                    | NAME
        '''
        if len(p) == 4 and p[2] == "as":
            # ensure import names / aliases are lower case
            if p[1].islower() and p[3].islower():
                p[0] = [p[1], p[3]]
        elif len(p) == 2 and p[1].islower():
            p[0] = p[1]

    def p_import_items_multiline(self, p):
        '''
        import_items_multiline : LPAR NEWLINE import_name NEWLINE RPAR
        '''

    # -- RULES --

    def p_rule(self, p):
        '''
        rule : NAME ASSIGN rhs terminator
             | NAME ASSIGN rhs
        '''
        p[0] = ('RULE', (p[1], p[3]))

    def p_rhs(self, p):
        '''
        rhs : factor NEWLINE VBAR rhs
            
            | factor VBAR rhs

            | factor NEWLINE
            | factor rhs
            | factor
        '''
        if len(p) == 5:
            p[0] = (p[3], (p[1], p[4]))
        elif len(p) == 4:
            p[0] = (p[2], (p[1], p[3]))
        elif len(p) == 3:
            p[0] = p[1]
            if isinstance(p[2], (list, tuple)):
                p[0] = (p[1], p[2])
            else:
                if p[2] != "\n":
                    p[0] = (p[1], p[2])
                else:
                    p[0] = p[1]
        elif len(p) == 2:
            p[0] = p[1]

    def p_factor(self, p):
        """
        factor : term OPTIONAL
               | term ASTERISK
               | term PLUS
               | term MINUS term
               | term
        """
        if len(p) == 4:
            p[0] = [p[2], [p[1], p[3]]]
        elif len(p) == 3:
            p[0] =  [p[2], p[1]]
        else:
            p[0] = p[1]

    def p_term_group(self, p):
        '''
        term_group : factor NEWLINE VBAR term_group
                   | factor VBAR term_group
                   | factor term_group
                   | factor
        '''
        if len(p) == 5:
            p[0] = (p[3], [p[1], p[4]])
        elif len(p) == 4:
            p[0] = (p[2], [p[1], p[3]])
        elif len(p) == 3:
            p[0] = ('CONCAT', (p[1], p[2]))
        elif len(p) == 2:
            p[0] = p[1]

    def p_term(self, p):
        """
        term : LPAR term_group RPAR
             | LSQB term_group RSQB
             | LBRACE term_group RBRACE
             | regex
             | string
             | name
        """
        if len(p) == 4:
            p[0] = (lexers.literal_type_map.get(p[1]), p[2])
        elif len(p) == 2:
            p[0] = p[1]

    def p_regex(self, p):
        '''
        regex : REGEX
        '''
        p[0] = ('REGEX', p[1])

    def p_string(self, p):
        '''
        string : STRING
        '''
        p[0] = ('STRING', p[1])

    def p_name(self, p):
        '''
        name : NAME
        '''
        p[0] = ('NAME', p[1])

    def p_terminator(self, p):
        '''
        terminator : SEMICOLON
                   | DOT
        '''
        p[0] = ('TERMINATOR', p[1])

    def find_column(self, lexdata, lexpos):
        line_start = lexdata.rfind("\n", 0, lexpos) + 1
        return (lexpos - line_start) + 1

    def p_error(self, t):
        if t is not None:
            print('error', self.parser.state)
            print(
                f"\033[91msyntax error: at line {t.lineno}, column {self.find_column(t.lexer.lexdata, t.lexpos)}, unexpected token {repr(t.value[0])}\033[0m"
            )
            exit(1)

    def parse(self, code, tracking=True):
        self.parser.parse(code, tracking=tracking)
        return self.ast



"""
ReBNF: Regexes for Extended Backus-Naur Form (EBNF)

Copyright (C) 2023 opsocket <opsocket@pm.me>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses>.
"""
from . import tokenize, parse, __ascii__, __version__
import argparse

__description__ = Rf"""

{__ascii__} v{__version__}

"""


def tabulate(data, headers=None):
    """
    Utility to generate the tokens table string
    """
    table_data = [headers] + data if headers else data
    # determine the maximum width of each column
    col_widths = [
        max(len(str(row[i])) for row in table_data) for i in range(len(table_data[0]))
    ]
    formatted_rows = []
    for row in table_data:
        formatted_row = [str(row[i]).ljust(col_widths[i]) for i in range(len(row))]
        formatted_rows.append(formatted_row)
    return "\n".join(" ".join(row) for row in formatted_rows)


def fmtok(tok):
    """
    Utility to print comprehensive token values
    """
    for c in "\n\t\r\f\b":
        if tok.value.count(c):
            return repr(tok.value)
    if tok.value.startswith("'"):
        return f'"{tok.value}"'
    else:
        return f"'{tok.value}'"


def main():
    parser = argparse.ArgumentParser(
        "rebnf",
        description=__description__,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    base_parser = argparse.ArgumentParser(add_help=False)
    base_parser.add_argument("filepath", type=str, help="path to the file to parse")

    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")

    # create the subparser for 'tokenize'
    tokenize_parser = subparsers.add_parser("tokenize", parents=[base_parser])
    tokenize_parser.add_argument(
        "-t", "--table", action="store_true", help="prints tokens as a table"
    )
    tokenize_parser.set_defaults(func=tokenize)

    # create the subparser for 'parse'
    parse_parser = subparsers.add_parser("parse", parents=[base_parser])
    parse_parser.add_argument(
        "-p", "--pretty", action="store_true", help="prints a pretty output"
    )
    parse_parser.set_defaults(func=parse)

    # parse command-line arguments
    args = parser.parse_args()

    # call appropriate function based on subcommand
    if hasattr(args, "func"):
        with open(args.filepath, "r") as code:
            if args.subcommand == "tokenize":
                tokens = args.func(code.read())
                print(
                    tabulate(
                        [[t.lineno, t.column, t.lexpos, t.type, fmtok(t)] for t in tokens],
                        headers=[
                            "LINE",
                            "COL",
                            "POS",
                            "TYPE",
                            "TOKEN",
                        ],
                    )
                )
            else:
                if args.pretty:
                    from pprint import pprint
                    pprint(args.func(code.read()))
                else:
                    print(args.func(code.read()))
    else:
        parser.print_help()


if __name__ == "__main__":
    main()
